/* 
 * This is free software released under the GPL license.
 * See the GNU GPL for details.
 *
 * (c) Juan-Mariano de Goyeneche. 1998, 1999.
 *
 */
 
/*
 *
 * V1.00 27/11/99
 * Modifications to work under RISC OS or unix, and to not fork ()
 * (c) Joseph Heenan, 1999.
 *
 */


/* Configuration parameters */
#define MAXLEN 1024
#define DELAY 5
#define TTL 2


#include <stdio.h>          /* printf(), snprintf() */
#include <stdlib.h>         /* strtol(), exit() */
#include <time.h>           /* time() */
#include "sys/types.h"
#include "sys/socket.h"     /* socket(), setsockopt(), bind(), recvfrom(), sendto() */
#include "sys/errno.h"      /* perror() */
#include "sys/ioctl.h"
#include "netinet/in.h"     /* IPPROTO_IP, sockaddr_in, htons(), htonl() */
#include "arpa/inet.h"      /* inet_addr() */
#include "string.h"         /* memset() */

#ifdef __riscos
#include "sys/byteorder.h"
#include "sys/select.h"
extern int close(int /*s*/); /* NetLib doesn't define this for us */
#else
#include "unistd.h"         /* gethostname() */
#endif


/* Sockets for sending and receiving. */
static int send_s = -1;
static int recv_s = -1;


/* registered with atexit() */
static void finalise(void)
{
  if (send_s != -1)
    close(send_s);

  if (recv_s != -1)
    close(recv_s);
}
    


/* Enable sleeping in taskwindows under RISC OS */
#ifdef __riscos
static void socket_twsleep(int s)
{
  int temp = 1;
  
  if (ioctl(s,FIOSLEEPTW, &temp)==-1)
  {
    perror("FIOSLEEPTW");
    exit(1);
  }
}

/* NetLib doesn't provide gethostname */
static int riscos_gethostname(char *name, size_t len)
{
  char *env = getenv("Inet$HostName");
  
  if (!env)
    return -1;
  
  name[0] = '\0';
  strncat(name, env, len - 1);
  return 0;
}
#define gethostname(name,len) riscos_gethostname(name,len)

#else

#define socket_twsleep(s)

#endif



int main(int argc, char* argv[])
{
  char nodename[40];
  u_char no = 0;
  u_int yes = 1;      /* Used with SO_REUSEADDR. In Linux both u_int */
                      /* and u_char are valid. */
  u_char ttl;
  struct sockaddr_in mcast_group;
  struct ip_mreq mreq;
  fd_set readfds;
  struct timeval timeout;
  time_t lastsend;

  if (argc<3 || argc>4)
  {
    fprintf(stderr, "Usage: %s mcast_group port [ttl]\n", argv[0]);
    exit(1);
  }

  if (gethostname(nodename, sizeof(nodename)-1) == -1 || !*nodename)
    strcpy(nodename, "[unknown]");

  ttl = argc == 4 ? atoi(argv[3]) : TTL;

  memset(&mcast_group, 0, sizeof(mcast_group));
  mcast_group.sin_family = AF_INET;
  mcast_group.sin_port = htons((unsigned short int)strtol(argv[2], NULL, 0));
  mcast_group.sin_addr.s_addr = inet_addr(argv[1]);
  
  atexit(finalise);
  
  if ( (send_s=socket(AF_INET, SOCK_DGRAM, 0)) == -1)
  {
    perror("send socket");
    exit(1);
  }
  
  socket_twsleep(send_s);

  /* Set non-blocking */
  if (ioctl(send_s, FIONBIO, &yes) == -1)
  {
    perror("ioctl");
    exit(1);
  }

  if (setsockopt(send_s, IPPROTO_IP, IP_MULTICAST_TTL, (char *)&ttl,
          sizeof(ttl)) < 0)
  {
    perror("ttl setsockopt");
    exit(1);
  }

  /* Disable Loop-back */
  if (setsockopt(send_s, IPPROTO_IP, IP_MULTICAST_LOOP, (char *)&no,
      sizeof(no)) < 0)
  {
    perror("loop setsockopt");
    exit(1);
  }

  if ( (recv_s=socket(AF_INET, SOCK_DGRAM, 0)) < 0)
  {
    perror("recv socket");
    exit(1);
  }

  socket_twsleep(recv_s);

  if (ioctl(recv_s, FIONBIO, &yes) == -1)
  {
    perror("ioctl");
    exit(1);
  }

  if (setsockopt(recv_s, SOL_SOCKET, SO_REUSEADDR, (char *)&yes, sizeof(yes)) < 0)
  {
    perror("reuseaddr setsockopt");
    exit(1);
  }

  if (bind(recv_s, (struct sockaddr*)&mcast_group, sizeof(mcast_group)) < 0)
  {
    perror ("bind");
    exit(1);
  }

  /* Tell the kernel we want to join that multicast group. */
  mreq.imr_multiaddr = mcast_group.sin_addr;
  mreq.imr_interface.s_addr = htonl(INADDR_ANY);
  if (setsockopt(recv_s, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&mreq, sizeof(mreq)) < 0)
  {
    perror ("add_membership setsockopt");
    exit(1);
  }
  
  lastsend = time(NULL);
  
  while (1)
  {
    FD_ZERO(&readfds);
    FD_SET(recv_s, &readfds);
    timerclear(&timeout);
    timeout.tv_sec = 1;
    
    switch (select(recv_s + 1, &readfds, NULL, NULL, &timeout))
    {
      
      case -1:
#ifndef __riscos
        if (errno != EINTR)
#endif
        {
          perror("select");
          exit(1);
        }
        break;
     
      case 0:
        /* timeout */
        break;
        
      default:
      {
        int n;
        char message[MAXLEN];
        struct sockaddr_in from;
        int len = sizeof(from);
        
        /* select successful, probably data to read */
        
        if ( (n=recvfrom(recv_s, message, MAXLEN-1, 0, 
                        (struct sockaddr*)&from, &len)) == -1 && errno != EWOULDBLOCK)
        {
          perror ("recv");
          exit(1);
        }
        message[n] = 0; /* null-terminate string */
        printf("%s: Received message from %s.\n", nodename,
                inet_ntoa(from.sin_addr));
        printf("\t%s\n", message);
      }
    } /* switch select */
    
    if (time(NULL) - lastsend >= DELAY)
    {
      /* timeout expired, send a packet */
      char message[MAXLEN];
      sprintf(message,  "Hi, I'm %s. Merry Christmas!\t(TTL==%d)", nodename, ttl);
      if (sendto(send_s, message, strlen(message), 0,
                  (struct sockaddr*)&mcast_group, 
                  sizeof(mcast_group)) < strlen (message))
      {
        perror("sendto");
        exit(1);
      }
      printf("Sending...\n");
      lastsend = time(NULL);
    }
  } /* forever */
  
  /* Not reached. */
}
